import logging
import matplotlib.pyplot as plt
import numpy as np
import os
from pathlib import Path
from pydrake.all import (
    Context,
    OutputPort,
    PortDataType,
    MultibodyPlant,
)
from pydrake.systems.primitives import VectorLogSink
from typing import List, Tuple


# Internal Imports
from brom_drake.PortWatcher.file_manager import PortWatcherFileManager
from brom_drake.PortWatcher.port_figure_arrangement import PortFigureArrangement
from brom_drake.PortWatcher.port_watcher_options import FigureNamingConvention, PortWatcherPlottingOptions
from brom_drake.utils.constants import SupportedLogger
from brom_drake.utils.type_checking import is_rigid_transform
from brom_drake.directories import DEFAULT_PLOT_DIR

class PortWatcherPlotter:
    """
    *Description*

    A plotter for the PortWatcher object. This is responsible for interpreting
    the data from the :py:class:`brom_drake.PortWatcher.PortWatcher` 
    
    TODO(Kwesi): Consider making this a dataclass
    """
    def __init__(
        self,
        port: OutputPort,
        python_logger: logging.Logger,
        file_manager: PortWatcherFileManager,
        plotting_options: PortWatcherPlottingOptions = PortWatcherPlottingOptions(),
    ):
        # Setup
        self.port = port
        self.plotting_options = plotting_options
        self.file_manager = file_manager
        self.python_logger = python_logger

    def compute_plot_shape(self, n_dims: int) -> Tuple[int, int]:
        """
        *Description*
        
        Computes the shape of the plot based on the data.

        *Parameters*
        
        n_dims: int
            The number of dimensions in the data.

        *Returns*
        
        n_rows: int
            The number of rows in the plot.

        n_columns: int
            The number of columns in the plot
        """
        if n_dims == 1:
            return 1, 1
        if n_dims == 2:
            return 1, 2

        if n_dims < 9:
            return 2, int(np.ceil(n_dims / 2.0))

        # Otherwise
        return 3, int(np.ceil(n_dims / 3.0))

    def add_to_python_report(self, message: str):
        """
        Description
        -----------
        Logs a message to the Python logger.
        :param message: A string with the message we want to send to the logs.
        :return:
        """
        self.python_logger.info(message)

    def add_warning_to_python_report(self, message: str):
        self.python_logger.warning(message)

    def data_dimension(self) -> int:
        """
        *Description*
        
        Returns the dimension of the data in the port.
        
        *Parameters*
        
        self : PortWatcherPlotter
            The PortWatcherPlotter object.

        *Returns*
        
        data_dim: int
            The dimension of the data in the port.
        """
        if self.port.get_data_type() == PortDataType.kVectorValued:
            return self.port.size()
        else:
            # If port contains RigidTransform, then the expected data dimension is 7.
            example_allocation = self.port.Allocate()
            example_value = example_allocation.get_value()
            if is_rigid_transform(example_value):
                return 7
            elif type(example_value) == bool: # if the output_value is a boolean
                return 1
        
        # Otherwise, raise an error
        raise ValueError(
            f"Port {self.port.get_name()} of system {self.port.get_system().get_name()} is not of the correct type for plotting."
        )

    def plot_logger_data(
        self,
        drake_vector_log: VectorLogSink,
        diagram_context: Context,
    ) -> Tuple[List[plt.Figure], List[List[plt.Axes]]]:
        """
        *Description*
        
        This function plots the data in the logger.

        *Parameters*
        
        self : PortWatcherPlotter
            The PortWatcherPlotter object.

        diagram_context : Context
            The context of the diagram.
        
        *Returns*
        
        Tuple[List[plt.Figure], List[List[plt.Axes]]]
            A tuple where:
            - the first element is a list of figures and 
            - the second element is a list of lists of axes.
        """
        # Setup
        python_logger = self.python_logger
        plotting_options = self.plotting_options
        system = self.port.get_system()

        # Get the log from the drake_vector_log
        temp_log = drake_vector_log.FindLog(diagram_context)

        log_times = temp_log.sample_times()
        log_data = temp_log.data()

        if (log_data.shape[1] == 0) or (log_data.shape[0] == 0):
            python_logger.warning(
                f"No data found for {system.get_name()} - Port {self.port.get_name()}! Skipping...")
            return None, None

        # Plot the data
        if plotting_options.plot_arrangement == PortFigureArrangement.OnePlotPerPort:
            fig, ax_list = self.plot_logger_data_subplots(log_times, log_data, drake_vector_log)
            return [fig], [ax_list]

        elif plotting_options.plot_arrangement == PortFigureArrangement.OnePlotPerDim:
            figs, ax_grid = [], []
            for port_index in range(self.port.size()):
                fig_ii = plt.figure()
                ax_ii = fig_ii.add_subplot(1, 1, 1)
                ax_ii.plot(
                    log_times, log_data[port_index, :]
                )

                # Add axis titles and labels
                ax_ii.set_xlabel("Time (s)")
                ax_ii.set_title(self.file_manager.name_of_data_at_index(port_index, self.port, drake_vector_log))

                # Save figures and axes to lists
                figs.append(fig_ii)
                ax_grid.append([ax_ii])

            return figs, ax_grid

        else:
            raise ValueError(
                f"Invalid plot arrangement: {plotting_options.plot_arrangement}."
            )

    def plot_logger_data_subplots(
        self,
        times: np.array,
        data: np.array,
        drake_vector_log: VectorLogSink,
    ):
        """
        *Description*
        
        This function plots the data in the logger.

        TODO(Kwesi): Consider adding refactoring this to remove dependecny on drake_vector_log

        *Parameters*
        
        self : PortWatcherPlotter
            The PortWatcherPlotter object.

        times : np.array
            The times at which the data was recorded.

        data : np.array
            The data that was recorded.
        
        *Returns*
        
        fig_out: plt.Figure
            The figure containing subplots that we use in the output.

        axes_list: list[plt.Axes]
            The list of axes.
        """
        # Setup
        n_dims = data.shape[0]

        # Plot the data
        n_rows, n_cols = self.compute_plot_shape(n_dims)

        self.add_to_python_report(
            f"Plotting {n_dims} dimensions in a {n_rows}x{n_cols} grid."
            )

        fig, ax_list = plt.subplots(n_rows, n_cols)

        if n_rows == 1 and n_cols == 1:
            ax_list.plot(times, data[0, :])
            ax_list.set_title(self.file_manager.name_of_data_at_index(0, self.port, drake_vector_log))

        elif (n_rows == 1) or (n_cols == 1):
            for dim_index in range(n_dims):
                ax_list[dim_index].plot(times, data[dim_index, :])
                ax_list[dim_index].set_title(
                    self.file_manager.name_of_data_at_index(dim_index, self.port, drake_vector_log),
                )

        else:
            for row_index in range(n_rows):
                for col_index in range(n_cols):

                    dim_index = n_cols * row_index + col_index

                    if dim_index >= n_dims:
                        fig.delaxes(ax_list[row_index, col_index])
                        continue

                    ax_list[row_index, col_index].plot(times, data[dim_index, :])
                    ax_list[row_index, col_index].set_title(self.file_manager.name_of_data_at_index(dim_index, self.port, drake_vector_log))

        return fig, ax_list

    def save_figures(
        self,
        vector_log_sink: VectorLogSink,
        diagram_context: Context,
        port_component_name: str|None = None,
    ):
        """
        *Description*
        
        This function saves the figures.

        .. note::

            TODO(kwesi): Make it so that this function computes names + directory structure based on plot arrangement.

        *Parameters*
        
        self : PortWatcherPlotter
            The PortWatcherPlotter object.

        diagram_context : Context
            The context of the diagram.
        """
        # Setup
        plotting_options = self.plotting_options
        figs, ax_list = self.plot_logger_data(vector_log_sink, diagram_context)

        # If no figures are returned, then return early!
        if figs is None:
            self.add_warning_to_python_report(
                f"No figures to save for {self.port.get_name()} of system {self.port.get_system().get_name()}; plot_logger_data was empty."
            )
            return

        if len(figs) == 0:
            self.add_warning_to_python_report(
                "Zero figures to save; plot_logger_data was empty."
            )
            return # Do nothing

        # Save the figures
        figure_paths = self.file_manager.compute_path_for_each_figure(
            output_port=self.port,
            associated_log_sink=vector_log_sink,
            port_component_name=port_component_name,
        )

        if len(figs) == 1:
            # Create directory for the plots, if it doesn't already exist
            first_figure_path = figure_paths[0]
            if not first_figure_path.parent.exists():
                first_figure_path.parent.mkdir(parents=True, exist_ok=True)

            # Save the single figure
            figs[0].savefig(first_figure_path, dpi=plotting_options.plot_dpi)
            plt.close(figs[0])
        else:
            # Plot each figure within this directory
            for ii, fig_ii in enumerate(figs):
                figure_path = figure_paths[ii]
                if not figure_path.parent.exists():
                    figure_path.parent.mkdir(parents=True, exist_ok=True)

                # Save figure to file
                fig_ii.savefig(figure_path, dpi=plotting_options.plot_dpi)
                plt.close(fig_ii)

        # Close all figures when done (this should be redundant?)
        # plt.close('all')

    def system_is_multibody_plant(self) -> bool:
        """
        *Description*
        
        Returns True if the system is a MultibodyPlant.
        
        *Parameters*
        
        self : PortWatcherPlotter
            The PortWatcherPlotter object.

        *Returns*
        
        is_plant: bool
            True if the system is a MultibodyPlant.
        """
        # Setup
        system = self.port.get_system()

        # Return
        return type(system) == MultibodyPlant


import networkx as nx
import numpy as np
from pydrake.common.eigen_geometry import Quaternion
from pydrake.common.value import AbstractValue
from pydrake.geometry import SceneGraph
from pydrake.math import RotationMatrix
from pydrake.multibody.inverse_kinematics import InverseKinematics
from pydrake.multibody.plant import MultibodyPlant
from pydrake.multibody.tree import ModelInstanceIndex
from pydrake.solvers import Solve, SolutionResult
from pydrake.systems.framework import LeafSystem, BasicVector, Context
from typing import Callable, Tuple

# Internal Imports
from brom_drake.utils.constants import MotionPlan, MotionPlanningResult


class PrototypicalPlannerSystem(LeafSystem):
    """
    Description
    -----------
    This function is meant to be a prototypical planner system.
    It will receive a planning algorithm and wrap that algorithm in a LeafSystem
    that works in Brom.
    """
    def __init__(
        self,
        plant: MultibodyPlant,
        scene_graph: SceneGraph,
        planning_algorithm: Callable[
            [np.ndarray, np.ndarray, Callable[[np.ndarray], bool]],
            Tuple[MotionPlan, int],
        ],
        robot_model_idx: ModelInstanceIndex = None,
        **kwargs,
    ):
        """
        Description
        -----------
        Constructor for the PrototypicalPlannerSystem.

        Arguments
        ---------
        plant : MultibodyPlant
            The plant that we are working with.
        scene_graph : SceneGraph
            The scene graph that we are working with.
        planning_algorithm : Callable[[np.ndarray, np.ndarray, Callable[[np.ndarray], bool]], Tuple[MotionPlan, int]]
            The planning algorithm that we are using.
            Note that the planning algorithm receives:
            1. The start configuration
            2. The goal configuration
            3. A collision checking function
            and it produces:
            1. The plan (can be an RRT tree or a numpy array of the list of configurations to visit)
            2. An optional index of the goal node within the plan. (Not needed if the plan is a numpy array)
        robot_model_idx : ModelInstanceIndex
            The robot model index that we are working with.
        """
        LeafSystem.__init__(self)

        # Setup
        self.plant, self.scene_graph = plant, scene_graph
        self.planning_algorithm = planning_algorithm
        self.robot_model_idx = robot_model_idx

        # Define Input Ports
        self.root_context = None  # NOTE: This should be assigned by the user before calling!
        self.plan = None
        self.create_input_ports()

        # Define Output Ports
        self.create_output_ports()

    def check_collision_in_config(
        self,
        q_model: np.ndarray,
    ) -> bool:
        """
        Description
        -----------
        This function checks for collisions in the robot's environment.

        Arguments
        ---------
        q_model : np.ndarray
            The model configuration that we are checking for collisions.
        """
        # Input Processing
        if self.root_context is None:
            raise ValueError("Plant context is not initialized yet!")

        if self.robot_model_idx is None:
            raise ValueError("Robot model index is not set yet!")

        # Setup
        plant_context = self.plant.GetMyMutableContextFromRoot(self.root_context)
        scene_graph_context = self.scene_graph.GetMyMutableContextFromRoot(self.root_context)
        q_init = self.plant.GetPositions(plant_context, self.robot_model_idx)

        # Set the configuration
        self.plant.SetPositions(
            plant_context,
            self.robot_model_idx,
            q_model
        )

        # Check for collisions using the query object
        query_object = self.scene_graph.get_query_output_port().Eval(scene_graph_context)
        # return query_object.HasCollisions()

        # Alternative method to check for collisions
        closest_points = query_object.ComputePointPairPenetration()
        eps0 = 1e-2
        collision_detected = False
        # inspector = self.scene_graph.model_inspector()
        for pair_ii in closest_points:
            if pair_ii.depth > eps0:
                collision_detected = True
                break
            
        # Otherwise return false (after resetting the configuration)
        self.plant.SetPositions(
            plant_context,
            self.robot_model_idx,
            q_init
        )
        return collision_detected

    def compute_plan_if_not_available(self, context: Context, output: AbstractValue):
        """
        Description
        -----------
        This function computes the plan if it is not available.

        Arguments
        ---------
        context : Context
            The context that we are working with.
        output : AbstractValue
            The output that will be produced by the output port
        """
        # Print
        if self.plan is None:
            # Compute plan
            q_start = self.GetInputPort("start_configuration").Eval(context)
            q_goal = self.GetInputPort("goal_configuration").Eval(context)
            if self.robot_model_idx is None:
                self.robot_model_idx = self.GetInputPort("robot_model_index").EvalAbstract(context).get_value()

            if self.root_context is None:
                raise ValueError("Plant context is not initialized yet!")

            self.root_context = self.root_context

            # Plan and extract path
            planning_result = self.planning_algorithm(
                q_start,
                q_goal,
                self.check_collision_in_config
            )

            self.plan = self.planning_result_to_array(planning_result)

        output.SetFrom(
            AbstractValue.Make(self.plan)
        )

    def create_input_ports(self):
        """
        Description:
            This function creates the input ports for the RRT planner.
        """
        # Setup
        N = len(
            self.plant.GetActuatedJointIndices(self.robot_model_idx)
        )

        # Define outputs
        self.DeclareVectorInputPort(
            "start_configuration",
            BasicVector(np.zeros((N,))),
        )
        self.DeclareVectorInputPort(
            "goal_configuration",
            BasicVector(np.zeros((N,))),
        )
        self.DeclareAbstractInputPort(
            "robot_model_index",
            AbstractValue.Make(ModelInstanceIndex(-1))
        )

    def create_output_ports(self):
        """
        Description:
            This function creates the output ports for the RRT planner.
        """
        # Setup

        # Create output for plan
        sample_plan = np.zeros((0, self.n_actuated_dof))
        self.DeclareAbstractOutputPort(
            "motion_plan",
            lambda: AbstractValue.Make(sample_plan),
            self.compute_plan_if_not_available
        )

        # Create output for plan readiness
        self.DeclareAbstractOutputPort(
            "plan_is_ready",
            lambda: AbstractValue.Make(False),
            self.GetPlanIsReady,
        )

    def define_pose_ik_problem(
        self,
        input_pose_vec: np.ndarray,
        eps0: float = 2.5e-2,
    ) -> InverseKinematics:
        """
        Description
        -----------
        Sets up the inverse kinematics problem for the start pose
        input to theis function.
        :return:
        """
        # Setup
        plant_context = self.plant.GetMyMutableContextFromRoot(self.root_context)

        # Create IK Problem
        ik_problem = InverseKinematics(self.plant)

        # Add Pose Target
        ik_problem.AddPositionConstraint(
            self.plant.world_frame(),
            input_pose_vec[:3],
            self.plant.GetFrameByName("ft_frame"),
            (- np.ones((3,)) * eps0).reshape((-1, 1)),
            (+ np.ones((3,)) * eps0).reshape((-1, 1)),
        )

        # Add Orientation Cost (not a hard constraint to help with feasibility)
        ik_problem.AddOrientationCost(
            self.plant.world_frame(),
            RotationMatrix(Quaternion(input_pose_vec[3:]).rotation()),
            self.plant.GetFrameByName("ft_frame"),
            RotationMatrix.Identity(),
            0.25,
        )

        return ik_problem

    @property
    def n_actuated_dof(self) -> int:
        """
        Description
        -----------
        This method uses the plant and the robot model idx
        to identify the numer of degrees of freedom we have for control.
        :return:
        """
        return self.plant.num_actuated_dofs()

    def GetPlanIsReady(self, context, output: AbstractValue):
        """
        Description:
            This function checks if the plan is ready.
        """
        output.SetFrom(
            AbstractValue.Make(self.plan is not None)
        )

    def planning_result_to_array(self, planning_result: MotionPlanningResult) -> np.ndarray:
        """
        Description
        -----------
        This function converts the planning result to an array.

        Arguments
        ---------
        planning_result : MotionPlanningResult
            The plan that we are converting to an array.
        """
        # Input Processing
        if isinstance(planning_result, tuple):
            plan_in, goal_node_index = planning_result

            if goal_node_index == -1:
                raise RuntimeError("No path found! Try increasing the number of iterations or checking your problem!")

        elif isinstance(planning_result, np.ndarray):
            plan_in = planning_result
            goal_node_index = -1 # Assume that the last configuration is the one we're trying to get to!

        elif isinstance(planning_result, nx.DiGraph):
            plan_in = planning_result
            all_node_ids = [int(node) for node in plan_in.nodes]
            goal_node_index = all_node_ids[-1] # Assume that the last configuration is the one we're trying to get to!

        else:
            raise ValueError(f"Unknown planning result type: \"{type(planning_result)}\"!")

        # Setup

        if type(plan_in) == np.ndarray:
            return plan_in
        elif type(plan_in) == nx.DiGraph:
            path = nx.shortest_path(
                plan_in,
                source=0,
                target=goal_node_index,
            )
            return np.array([plan_in.nodes[node]['q'] for node in path])
        else:
            raise ValueError(f"Unknown plan type: \"{type(plan_in)}\"!")

    def set_internal_root_context(self, root_context_in: Context):
        self.root_context = root_context_in


from dataclasses import dataclass

@dataclass
class NetworkXFSMConfig:
    """
    Configuration for the NetworkXFSM system.
    
    Attributes:
        name (str): Name of the FSM.
        initial_state (str): Initial state of the FSM.
        states (list): List of states in the FSM.
        transitions (dict): Mapping of transitions between states.
    """
    name: str = "NetworkXFSM"
    log_file_name: str = "network_fsm.log"
    show_logs_in_terminal: bool = True


from pydrake.multibody.plant import MultibodyPlant
from pydrake.multibody.tree import ModelInstanceIndex, BodyIndex
from typing import List

def get_all_associated_body_indices_in_plant(
    plant: MultibodyPlant,
) -> List[BodyIndex]:
    """
    *Description*
    
    This method returns all the bodies in the given model instance.
    """
    # Iterate through all model instances and collect body indices
    body_indices_out = []
    for model_instance_index in range(plant.num_model_instances()):
        model_body_indices = plant.GetBodyIndices(ModelInstanceIndex(model_instance_index))
        body_indices_out.extend(model_body_indices)
    
    return body_indices_out



from brom_drake.utils.leaf_systems.network_fsm.fsm_edge_definition import FSMEdgeDefinition
from brom_drake.utils.leaf_systems.network_fsm.fsm_output_definition import FSMOutputDefinition
from brom_drake.utils.leaf_systems.network_fsm.fsm_transition_condition import (
    FSMTransitionCondition,
    FSMTransitionConditionType,
)
from brom_drake.utils.leaf_systems.network_fsm.networkx_fsm import NetworkXFSM
from brom_drake.utils.leaf_systems.network_fsm.config import NetworkXFSMConfig

__all__ = [
    "FSMEdgeDefinition",
    "FSMOutputDefinition",
    "FSMTransitionCondition",
    "FSMTransitionConditionType",
    "NetworkXFSM",
    "NetworkXFSMConfig",
]


from brom_drake.productions.debug.grasping.show_me_this_static_grasp.production import (
    ShowMeThisStaticGrasp,
)
from .show_me_this_static_grasp.config import Configuration as ShowMeThisStaticGraspConfiguration
from .attempt_grasp.with_static_wrist.production import AttemptGraspWithStaticWrist
from .attempt_grasp.with_static_wrist.config import Configuration as AttemptGraspWithStaticWristConfiguration

__all__ = [
    "ShowMeThisStaticGrasp",
    "ShowMeThisStaticGraspConfiguration",
    "AttemptGraspWithStaticWrist",
    "AttemptGraspWithStaticWristConfiguration",
]


from dataclasses import dataclass
import numpy as np
from typing import Union

@dataclass
class FSMOutputDefinition:
    """
    Dataclass to represent the output of a state in a finite state machine.
    """
    output_port_name: str
    output_port_value: Union[bool, float, np.ndarray, str]

    #TODO(kwesi): Eventually it would be cool to add support for
    #             output_port_value to be a function that can be called.
    #             This would allow for more dynamic outputs and can be
    #             checked with the "callable" method.
